"""
@Author: Xin Cao, Xiangning Chu, University of Colorado Boulder
The vector and contour map plots are implemented for initial investigation.
"""
import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
from matplotlib.colors import TwoSlopeNorm, Normalize, SymLogNorm, NoNorm, LogNorm #, CenteredNorm
from .config import CONFIG
from itertools import chain
from matplotlib.path import Path
from pyspedas.secs import read_data_files
from datetime import datetime, timedelta
from pyspedas.utilities.dailynames import dailynames
from mpl_toolkits.basemap.solar import daynight_terminator
import logging
#os.environ['PROJ_LIB'] = '/Users/cao/anaconda3/envs/secs/share/proj'
#if an error about "dedent" occurs, downgrade the matplotlib version to 3.2.0 by using "pip install -U matplotlib==3.2"
try:
    from mpl_toolkits.basemap import Basemap
except ImportError:
    logging.info('Error importing Basemap; installation instructions can be found at: https://matplotlib.org/basemap/users/installing.html or https://anaconda.org/anaconda/basemap')


class CenteredNorm(Normalize):
    """
    This function is directly contributed from the latest matplotlib version,
    but not included in the 3.2.0 version, so is duplicated here.
    """
    def __init__(self, vcenter=0, halfrange=None, clip=False):
        """
        Normalize symmetrical data around a center (0 by default).
        Unlike `TwoSlopeNorm`, `CenteredNorm` applies an equal rate of change
        around the center.
        Useful when mapping symmetrical data around a conceptual center
        e.g., data that range from -2 to 4, with 0 as the midpoint, and
        with equal rates of change around that midpoint.
        Parameters
        ----------
        vcenter : float, default: 0
            The data value that defines ``0.5`` in the normalization.
        halfrange : float, optional
            The range of data values that defines a range of ``0.5`` in the
            normalization, so that *vcenter* - *halfrange* is ``0.0`` and
            *vcenter* + *halfrange* is ``1.0`` in the normalization.
            Defaults to the largest absolute difference to *vcenter* for
            the values in the dataset.
        Examples
        --------
        This maps data values -2 to 0.25, 0 to 0.5, and 4 to 1.0
        (assuming equal rates of change above and below 0.0):
            >>> import matplotlib.colors as mcolors
            >>> norm = mcolors.CenteredNorm(halfrange=4.0)
            >>> data = [-2., 0., 4.]
            >>> norm(data)
            array([0.25, 0.5 , 1.  ])
        """
        super().__init__(vmin=None, vmax=None, clip=clip)
        self._vcenter = vcenter
        # calling the halfrange setter to set vmin and vmax
        self.halfrange = halfrange

    def _set_vmin_vmax(self):
        """
        Set *vmin* and *vmax* based on *vcenter* and *halfrange*.
        """
        self.vmax = self._vcenter + self._halfrange
        self.vmin = self._vcenter - self._halfrange

    def autoscale(self, A):
        """
        Set *halfrange* to ``max(abs(A-vcenter))``, then set *vmin* and *vmax*.
        """
        A = np.asanyarray(A)
        self._halfrange = max(self._vcenter-A.min(),
                              A.max()-self._vcenter)
        self._set_vmin_vmax()

    def autoscale_None(self, A):
        """Set *vmin* and *vmax*."""
        A = np.asanyarray(A)
        if self._halfrange is None and A.size:
            self.autoscale(A)

    @property
    def vcenter(self):
        return self._vcenter

    @vcenter.setter
    def vcenter(self, vcenter):
        if vcenter != self._vcenter:
            self._vcenter = vcenter
            self._changed()
        if self.vmax is not None:
            # recompute halfrange assuming vmin and vmax represent
            # min and max of data
            self._halfrange = max(self._vcenter-self.vmin,
                                  self.vmax-self._vcenter)
            self._set_vmin_vmax()

    @property
    def halfrange(self):
        return self._halfrange

    @halfrange.setter
    def halfrange(self, halfrange):
        if halfrange is None:
            self._halfrange = None
            self.vmin = None
            self.vmax = None
        else:
            self._halfrange = abs(halfrange)

    def __call__(self, value, clip=None):
        if self._halfrange is not None:
            # enforce symmetry, reset vmin and vmax
            self._set_vmin_vmax()
        return super().__call__(value, clip=clip)


def draw_map(m, scale=0.2):
    """
    This function is for data visualization with geographic map.
    Parameter m: basemap object
    """
    # draw a shaded-relief image
    m.shadedrelief(scale=scale)

    # lats and longs are returned as a dictionary
    lats = m.drawparallels(np.linspace(-90, 90, 13))
    lons = m.drawmeridians(np.linspace(-180, 180, 13))

    # keys contain the plt.Line2D instances
    lat_lines = chain(*(tup[1][0] for tup in lats.items()))
    lon_lines = chain(*(tup[1][0] for tup in lons.items()))
    all_lines = chain(lat_lines, lon_lines)

    # cycle through these lines and set the desired style
    for line in all_lines:
        line.set(linestyle='-', alpha=0.3, color='w')
    return


def noon_midnight_meridian(dtime=None, delta=0.25):
    """
    This function calculates the longtitude and latitude of the noon-midnight meridian
    based on a given UTC time.
    :param dtime: the str of UTC time, which is of the format of 'year-month-day/hour:minute:second'
    :param delta: float or int, in degree, which is the interval of neighbor points on the meridian.
    :return: dictionary with keys: 'lons_noonmidnight', 'lats_noonmidnight', 'lons_noon', 'lats_noon'
                                    'lons_midnight', 'lats_midnight'
                        with values of numpy arrays
    """
    # method2:
    n_interval = 360 / delta + 1
    ni_half = int(np.floor(n_interval / 2))
    ni_otherhalf = int(n_interval - ni_half)

    time_current_UTC = datetime.strptime(dtime, '%Y-%m-%d/%H:%M:%S')
    dtime_noon = dtime[0:11] + '12:00:00'
    # print('dtime_noon: ', dtime_noon)
    time_noon = datetime.strptime(dtime_noon, '%Y-%m-%d/%H:%M:%S')
    time_diff = time_noon - time_current_UTC
    diff_in_hours = time_diff.total_seconds() / 3600  # within [-12,12] hours due to the same day.
    if diff_in_hours == 0:
        lons_latmax = 0  # current UTC time is just at noon
        lons_latmin = 180  # midnight longitude
    elif diff_in_hours > 0:
        lons_latmax = 0 + 15 * diff_in_hours  # longitude for noon line
        lons_latmin = lons_latmax - 180  # longitude for midnight line
    elif diff_in_hours < 0:
        lons_latmax = 0 - 15 * diff_in_hours  # longitude for noon line
        lons_latmin = lons_latmax + 180  # longitude for midnight line
    #
    lons_max_arr = np.full((1, ni_half), lons_latmax)  # for noon line
    lats_max_arr = np.linspace(-90, 90, ni_half)  # for noon line

    lons_min_arr = np.full((1, ni_otherhalf), lons_latmin)  # for midnight line
    lats_min_arr = np.linspace(90, -90, ni_otherhalf)  # for midnight line

    lons_arr = np.concatenate((lons_max_arr, lons_min_arr), axis=None)
    lats_arr = np.concatenate((lats_max_arr, lats_min_arr), axis=None)
    lons_nm, lats_nm = lons_arr, lats_arr  # the whole noon-midnight circle

    lons_n, lats_n = lons_max_arr[0], lats_max_arr  # the noon semi-circle
    lons_m, lats_m = lons_min_arr[0], lats_min_arr  # the midnight semi-circle

    noon_midnight = {'lons_noonmidnight': lons_nm, 'lats_noonmidnight': lats_nm,
                     'lons_noon': lons_n, 'lats_noon': lats_n,
                     'lons_midnight': lons_m, 'lats_midnight': lats_m}
    return noon_midnight


def _make_EICS_plots(dtime=None, vplot_sized=False, contour_den=8, s_loc=False, quiver_scale=30, pred = False, pred_var = None, plot_title_addon = None, colormaplimit=None, EICS_component='Jy'):
    """
    @Parameter: dtime input as a string
    @Parameter: s_loc input as a bool, which means the locations of the virtual stations.
    """
    dtype = 'EICS'

    if not os.path.exists(CONFIG['plots_dir']):
        os.makedirs(CONFIG['plots_dir'])
    dtime_range = [dtime, dtime]
    pathformat_prefix = dtype + '/%Y/%m/'
    pathformat_unzipped = pathformat_prefix + '%d/' + dtype + '%Y%m%d_%H%M%S.dat'
    filename_unzipped = dailynames(file_format=pathformat_unzipped, trange=dtime_range, res=10)
    out_files_unzipped = [CONFIG['local_data_dir'] + rf_res for rf_res in filename_unzipped]
    Data_Days_time = read_data_files(out_files=out_files_unzipped, dtype=dtype, out_type='df')

    if pred == True: # XC
        obs_var = Data_Days_time['Jy']
        print(np.size(pred_var), np.size(Data_Days_time['Jy']), np.shape(pred_var), np.shape(Data_Days_time['Jy']))
        if np.size(pred_var) != np.size(Data_Days_time['Jy']):
            raise ValueError('np.size(pred_var) != np.size(Data_Days_time[Jy]) !')
        #if np.shape(pred_var) != np.shape(Data_Days_time['Jy']):
        #    raise ValueError('np.shape(pred_var) != np.shape(Data_Days_time[Jy]) !')
        pred_var = np.reshape(pred_var, np.shape(Data_Days_time['Jy']))
        if np.shape(pred_var) == np.shape(Data_Days_time['Jy']):
            print(' pred_var has the same size and shape with Data_Days_time[Jy] !')
        Data_Days_time['Jy'] = pred_var

    J_comp = Data_Days_time['Jy']
    Jc_max, Jc_min = J_comp.max(), J_comp.min()
    Jcm_abs = max(abs(Jc_max), abs(Jc_min))
    contour_density = np.linspace(-Jcm_abs, Jcm_abs, num=contour_den)
    tp = dtime
    datetime_tp = tp[0:4] + tp[5:7] + tp[8:10] + '_' + tp[11:13] + tp[14:16] + tp[17:19]
    lon = Data_Days_time['longitude']
    lat = Data_Days_time['latitude']
    Jx = Data_Days_time['Jx']  # Note: positive is Northward
    Jy = Data_Days_time['Jy']  # Note: positive is Eastward

    # plot 1:
    # plot map ground (North hemisphere)
    fig1 = plt.figure(figsize=(8, 8))
    ax1 = plt.gca()

    m = Basemap(projection='lcc', resolution='c',
                width=8E6, height=8E6,
                lat_0=60, lon_0=-100)
    # draw coastlines, country boundaries, fill continents.
    m.drawcoastlines(linewidth=0.25)
    m.drawcountries(linewidth=0.25)
    m.fillcontinents(color='None', lake_color='None')
    # draw the edge of the map projection region (the projection limb)
    m.drawmapboundary(fill_color=None)
    # m.drawgreatcircle(-100,0,0,90)
    m.drawlsmask()
    # m.bluemarble()
    m.shadedrelief()
    # draw parallels and meridians.
    # label parallels on right and top
    # meridians on bottom and left
    parallels = np.arange(0., 81, 10.)
    # labels = [left,right,top,bottom]
    m.drawparallels(parallels, labels=[False, True, True, False])
    meridians = np.arange(10., 351., 20.)
    m.drawmeridians(meridians, labels=[True, False, False, True])
    date_nightshade = datetime.strptime(dtime, '%Y-%m-%d/%H:%M:%S')
    m.nightshade(date=date_nightshade)
    draw_map(m)

    # plot vector field:
    print('lon before: ', lon)
    lon = lon.to_numpy()
    lat = lat.to_numpy()
    print('lon after:  ', lon)
    Jx = Jx.to_numpy()  # Note: positive is Northward
    Jy = Jy.to_numpy()  # Note: positive is Eastward
    Jx_uni = Jx / np.sqrt(Jx ** 2 + Jy ** 2)
    Jy_uni = Jy / np.sqrt(Jx ** 2 + Jy ** 2)
    n = -2
    color = np.sqrt(((Jx_uni - n) / 2) ** 2 + ((Jy_uni - n) / 2) ** 2)
    if vplot_sized == False:
        qv = m.quiver(lon, lat, Jx_uni, Jy_uni, color, headlength=7, latlon=True,
                      cmap='GnBu')  # autumn_r #, color=cm(norm(o)))#, cmap = 'jet')
        plt.colorbar()
    else:
        Jy_rot, Jx_rot, x, y = m.rotate_vector(Jy, Jx, lon, lat, returnxy=True)
        qv = m.quiver(lon, lat, Jy_rot, Jx_rot, headlength=7, latlon=True, scale_units='dots',
                      scale=quiver_scale)  # , transform='lcc')

        qk = ax1.quiverkey(qv, 0.3, -0.1, 100, r'$100 \ mA/m$', labelpos='E', coordinates='data')  # figure


    plt.title(label='EICS ' + tp, fontsize=20, color="black", pad=20)
    plt.tight_layout()
    plt.savefig(CONFIG['plots_dir'] + 'EICS' + '_vector_' + date_nightshade.strftime('%Y%m%d%H%M%S') + '.jpeg')
    plt.show()

    # plot 2: contour plot
    # plot map ground (North hemisphere)
    fig2 = plt.figure(figsize=(8, 8))
    ax2 = plt.gca()
    m = Basemap(projection='lcc', resolution='c',
                width=8E6, height=8E6,
                lat_0=60, lon_0=-100)
    # draw coastlines, country boundaries, fill continents.
    m.drawcoastlines(linewidth=0.25)
    m.drawcountries(linewidth=0.25)
    m.fillcontinents(color='None', lake_color='None')
    # draw the edge of the map projection region (the projection limb)
    m.drawmapboundary(fill_color=None)
    m.drawlsmask()
    m.shadedrelief()
    # draw parallels and meridians.
    # label parallels on right and top
    # meridians on bottom and left
    parallels = np.arange(0., 81, 10.)
    m.drawparallels(parallels, labels=[False, True, True, False])
    meridians = np.arange(10., 351., 20.)
    m.drawmeridians(meridians, labels=[True, False, False, True])

    date_nightshade = datetime.strptime(dtime, '%Y-%m-%d/%H:%M:%S')
    # m.nightshade(date=date_nightshade, alpha = 0.0)

    delta = 0.25
    lons_dd, lats_dd, tau, dec = daynight_terminator(date_nightshade, delta, m.lonmin, m.lonmax)
    xy = [lons_dd, lats_dd]
    xy = np.array(xy)
    xb, yb = xy[0], xy[1]
    m.plot(xb, yb, marker=None, color='m', latlon=True)  # for dawn-dusk circle line

    # Plot the noon-midnight line.
    n_interval = len(lons_dd)
    ni_half = int(np.floor(len(lons_dd) / 2))
    ni_otherhalf = n_interval - ni_half


    noon_midnight = noon_midnight_meridian(dtime, delta)

    m.plot(noon_midnight['lons_noon'], noon_midnight['lats_noon'], marker=None, color='deepskyblue',
           latlon=True)  # noon semi-circle

    m.plot(noon_midnight['lons_midnight'], noon_midnight['lats_midnight'], marker=None, color='k',
           latlon=True)  # midnight semi-circle

    draw_map(m)

    Jy_log = Jy / np.abs(Jy) * np.log10(np.abs(Jy))
    Jy_log = Jy / np.abs(Jy) * np.log10(np.abs(Jy)) ##
    if colormaplimit == None:
        norm_cb = CenteredNorm()
    else:
        norm_cb = CenteredNorm(halfrange=colormaplimit)
    # norm_cb = NoNorm()
    # norm_cb = CenteredNorm(vmin=Jy.min(), vcenter=0, vmax=Jy.max())
    # use Jy for the contour map, not Jy_rot.
    if EICS_component == 'Jy':
        ctrf = m.contourf(lon, lat, Jy, contour_density, latlon=True, tri=True, cmap='jet_r', norm=norm_cb)
    elif EICS_component == 'Jx':
        ctrf = m.contourf(lon, lat, Jx, contour_density, latlon=True, tri=True, cmap='jet_r', norm=norm_cb)
    ##ctrf = m.contourf(lon, lat, Jy, contour_density, latlon=True, tri=True, cmap='jet_r', norm=norm_cb)
    # -------------
    if s_loc:
        m.scatter(lon, lat, latlon=True, marker='*', c='black')
    FAC_loc = False
    if FAC_loc: # for 2008-02-29 and 2008-02-22 events' field aligned current locations:
        import pandas as pd
        event_df_a = pd.read_csv('/Users/xica3495/Downloads/event_locs/scwsnapshot_step1_nosubSQ_V2.txt', delimiter="\s+")
        event_df_b = pd.read_csv('/Users/xica3495/Downloads/event_locs/scwsnapshot_step2_nosubSQ_V2.txt', delimiter="\s+")
        event_df_c = pd.read_csv('/Users/xica3495/Downloads/event_locs/scwsnapshot_step3_nosubSQ_V2.txt', delimiter="\s+")
        event_df_d = pd.read_csv('/Users/xica3495/Downloads/event_locs/scwsnapshot_step4_nosubSQ_V2.txt', delimiter="\s+")
        #print('cp00!')
        #print(dtime)
        #print(event_df.loc[0,'time'])
        #print(dtime==event_df.loc[0,'time'])
        row_event_df_a = event_df_a[event_df_a['time'] == dtime]
        row_event_df_b = event_df_b[event_df_b['time'] == dtime]
        row_event_df_c = event_df_c[event_df_c['time'] == dtime]
        row_event_df_d = event_df_d[event_df_d['time'] == dtime]
        #print(row_event_df)
        #print(row_event_df['lat1_geo'].to_numpy()[0], type(row_event_df['lat1_geo'].to_numpy()[0]))
        lon1_ev_a = row_event_df_a['lon1_geo'].to_numpy()[0]
        lon2_ev_a = row_event_df_a['lon2_geo'].to_numpy()[0]
        lat1_ev_a = row_event_df_a['lat1_geo'].to_numpy()[0]
        lat2_ev_a = row_event_df_a['lat2_geo'].to_numpy()[0]

        lon1_ev_b = row_event_df_b['lon1_geo'].to_numpy()[0]
        lon2_ev_b = row_event_df_b['lon2_geo'].to_numpy()[0]
        lat1_ev_b = row_event_df_b['lat1_geo'].to_numpy()[0]
        lat2_ev_b = row_event_df_b['lat2_geo'].to_numpy()[0]

        lon1_ev_c = row_event_df_c['lon1_geo'].to_numpy()[0]
        lon2_ev_c = row_event_df_c['lon2_geo'].to_numpy()[0]
        lat1_ev_c = row_event_df_c['lat1_geo'].to_numpy()[0]
        lat2_ev_c = row_event_df_c['lat2_geo'].to_numpy()[0]

        lon1_ev_d = row_event_df_d['lon1_geo'].to_numpy()[0]
        lon2_ev_d = row_event_df_d['lon2_geo'].to_numpy()[0]
        lat1_ev_d = row_event_df_d['lat1_geo'].to_numpy()[0]
        lat2_ev_d = row_event_df_d['lat2_geo'].to_numpy()[0]


        m.scatter(lon1_ev_a, lat1_ev_a, latlon=True, marker='o', s=20, c='red')
        m.scatter(lon2_ev_a, lat2_ev_a, latlon=True, marker='o', s=20, c='blue')

        m.scatter(lon1_ev_b, lat1_ev_b, latlon=True, marker='*', s=20, c='red')
        m.scatter(lon2_ev_b, lat2_ev_b, latlon=True, marker='*', s=20, c='blue')

        m.scatter(lon1_ev_c, lat1_ev_c, latlon=True, marker='^', s=20, c='red')
        m.scatter(lon2_ev_c, lat2_ev_c, latlon=True, marker='^', s=20, c='blue')

        m.scatter(lon1_ev_d, lat1_ev_d, latlon=True, marker='D', s=20, c='red')
        m.scatter(lon2_ev_d, lat2_ev_d, latlon=True, marker='D', s=20, c='blue')

        #print(event_df.loc[0,:])
        #print(event_df.loc[1,:])
        #print(event_df.loc[2,['lat1_geo', 'lat2_geo']])
        #print(event_df['lon2_sm'])
        print('cp01!')
    # -------------
    cb = m.colorbar(matplotlib.cm.ScalarMappable(norm=norm_cb, cmap='jet_r'), pad='15%')
    if EICS_component == 'Jy':
        cb.set_label(r'$\mathit{J}_y \  (mA/m)$')
    elif EICS_component == 'Jx':
        cb.set_label(r'$\mathit{J}_x \  (mA/m)$')
    ax_cb = cb.ax
    text = ax_cb.yaxis.label
    font_cb = matplotlib.font_manager.FontProperties(family='times new roman', style='italic', size=20)
    text.set_font_properties(font_cb)
    if pred == True:
        if plot_title_addon != None:
            plt.title(label= plot_title_addon + ' Predicted EICS ' + tp, fontsize=20, color="black", pad=20)
        else:
            plt.title(label= 'Predicted EICS ' + tp, fontsize=20, color="black", pad=20)
    else:
        plt.title(label='EICS ' + tp, fontsize=20, color="black", pad=20)
    plt.tight_layout()
    if pred == True:
        if EICS_component == 'Jy':
            plt.savefig(CONFIG['plots_dir'] + 'EICS_Jy' + '_pred' + '_contour_' + date_nightshade.strftime('%Y%m%d%H%M%S') + '.jpeg')
        elif EICS_component == 'Jx':
            plt.savefig(CONFIG['plots_dir'] + 'EICS_Jx' + '_pred' + '_contour_' + date_nightshade.strftime(
                '%Y%m%d%H%M%S') + '.jpeg')
    else:
        if EICS_component == 'Jy':
            plt.savefig(CONFIG['plots_dir'] + 'EICS_Jy' + '_contour_' + date_nightshade.strftime('%Y%m%d%H%M%S') + '.jpeg')
        elif EICS_component == 'Jx':
            plt.savefig(CONFIG['plots_dir'] + 'EICS_Jx' + '_contour_' + date_nightshade.strftime('%Y%m%d%H%M%S') + '.jpeg')

    plt.show()

    # plot the visual grids:

    fig3 = plt.figure(figsize=(8, 8))
    ax3 = plt.gca()
    m = Basemap(projection='lcc', resolution='c',
                width=8E6, height=8E6,
                lat_0=60, lon_0=-100)
    # draw coastlines, country boundaries, fill continents.
    m.drawcoastlines(linewidth=0.25)
    m.drawcountries(linewidth=0.25)
    m.fillcontinents(color='None', lake_color='None')
    # draw the edge of the map projection region (the projection limb)
    m.drawmapboundary(fill_color=None)
    m.drawlsmask()
    m.shadedrelief()
    # draw parallels and meridians.
    # label parallels on right and top
    # meridians on bottom and left
    parallels = np.arange(0., 81, 10.)
    m.drawparallels(parallels, labels=[False, True, True, False])
    meridians = np.arange(10., 351., 20.)
    m.drawmeridians(meridians, labels=[True, False, False, True])

    date_nightshade = datetime.strptime(dtime, '%Y-%m-%d/%H:%M:%S')
    # m.nightshade(date=date_nightshade, alpha = 0.0)

    delta = 0.25
    lons_dd, lats_dd, tau, dec = daynight_terminator(date_nightshade, delta, m.lonmin, m.lonmax)
    xy = [lons_dd, lats_dd]
    xy = np.array(xy)
    xb, yb = xy[0], xy[1]
    m.plot(xb, yb, marker=None, color='m', latlon=True)  # for dawn-dusk circle line

    # Plot the noon-midnight line.
    n_interval = len(lons_dd)
    ni_half = int(np.floor(len(lons_dd) / 2))
    ni_otherhalf = n_interval - ni_half

    noon_midnight = noon_midnight_meridian(dtime, delta)

    m.plot(noon_midnight['lons_noon'], noon_midnight['lats_noon'], marker=None, color='deepskyblue',
           latlon=True)  # noon semi-circle

    m.plot(noon_midnight['lons_midnight'], noon_midnight['lats_midnight'], marker=None, color='k',
           latlon=True)  # midnight semi-circle

    draw_map(m)

    # -------------
    VG_labels = np.linspace(0, len(lon)-1, len(lon), dtype=int)
    m.scatter(lon, lat, latlon=True, marker='*', c='black')

    x_shift, y_shift = (-5, 5)
    for i, vg_label in enumerate(VG_labels):
        lon_tmp, lat_tmp = m(lon[i], lat[i])
        plt.annotate(str(vg_label), xy=(lon_tmp, lat_tmp), xycoords='data', xytext=(x_shift, y_shift), textcoords='offset points', color='r', fontsize=6) #xy=(lon[i], lat[i])
    # -------------
    font_cb = matplotlib.font_manager.FontProperties(family='times new roman', style='italic', size=20)
    text.set_font_properties(font_cb)
    plt.title(label='EICS visual grids' + tp, fontsize=20, color="black", pad=20)
    plt.tight_layout()
    plt.savefig(CONFIG['plots_dir'] + 'EICS' + '_VisualGrids_' + date_nightshade.strftime('%Y%m%d%H%M%S') + '.pdf')
    plt.show()



    print('EICS plots completed!')
    return


def _make_SECS_plots(data=None, dtime=None, contour_den=8, s_loc=False):
    """
    @Parameter: dtime input as a string
    @Parameter: s_loc input as a bool, which means the locations of the virtual stations.
    """
    dtype = 'SECS'

    if not os.path.exists(CONFIG['plots_dir']):
        os.makedirs(CONFIG['plots_dir'])
    dtime_range = [dtime, dtime]
    pathformat_prefix = dtype + '/%Y/%m/'
    pathformat_unzipped = pathformat_prefix + '%d/' + dtype + '%Y%m%d_%H%M%S.dat'
    filename_unzipped = dailynames(file_format=pathformat_unzipped, trange=dtime_range, res=10)
    out_files_unzipped = [CONFIG['local_data_dir'] + rf_res for rf_res in filename_unzipped]
    Data_Days_time = read_data_files(out_files=out_files_unzipped, dtype=dtype, out_type='df')

    J_comp = Data_Days_time['J']
    Jc_max, Jc_min = J_comp.max(), J_comp.min()
    Jcm_abs = max(abs(Jc_max), abs(Jc_min))
    contour_density = np.linspace(-Jcm_abs, Jcm_abs, num=contour_den)
    tp = dtime
    datetime_tp = tp[0:4] + tp[5:7] + tp[8:10] + '_' + tp[11:13] + tp[14:16] + tp[17:19]
    lon = Data_Days_time['longitude']
    lat = Data_Days_time['latitude']
    J = Data_Days_time['J']

    lon = lon.to_numpy()
    lat = lat.to_numpy()
    J = J.to_numpy()

    # plot 1: contour plot
    # plot map ground (North hemisphere)
    fig = plt.figure(figsize=(8, 8))
    m = Basemap(projection='lcc', resolution='c',
                width=8E6, height=8E6,
                lat_0=60, lon_0=-100)
    # draw coastlines, country boundaries, fill continents.
    m.drawcoastlines(linewidth=0.25)
    m.drawcountries(linewidth=0.25)
    m.fillcontinents(color='None', lake_color='None')
    # draw the edge of the map projection region (the projection limb)
    m.drawmapboundary(fill_color=None)
    m.drawlsmask()
    m.shadedrelief()
    # draw parallels and meridians.
    # label parallels on right and top
    # meridians on bottom and left
    parallels = np.arange(0., 81, 10.)
    m.drawparallels(parallels, labels=[False, True, True, False])
    meridians = np.arange(10., 351., 20.)
    m.drawmeridians(meridians, labels=[True, False, False, True])

    date_nightshade = datetime.strptime(dtime, '%Y-%m-%d/%H:%M:%S')
    delta = 0.25
    lons_dd, lats_dd, tau, dec = daynight_terminator(date_nightshade, delta, m.lonmin, m.lonmax)
    xy = [lons_dd, lats_dd]
    xy = np.array(xy)
    xb, yb = xy[0], xy[1]
    m.plot(xb, yb, marker=None, color='b', latlon=True)

    # Plot the noon-midnight line.
    n_interval = len(lons_dd)
    ni_half = int(np.floor(len(lons_dd) / 2))
    ni_otherhalf = n_interval - ni_half

    noon_midnight = noon_midnight_meridian(dtime, delta)

    m.plot(noon_midnight['lons_noon'], noon_midnight['lats_noon'], marker=None, color='deepskyblue',
           latlon=True)  # noon semi-circle

    m.plot(noon_midnight['lons_midnight'], noon_midnight['lats_midnight'], marker=None, color='k',
           latlon=True)  # midnight semi-circle

    draw_map(m)

    norm_cb = CenteredNorm()
    ctrf = m.contourf(lon, lat, J, contour_density, latlon=True, tri=True, cmap=plt.get_cmap('seismic', 20),
                      norm=norm_cb)
    if s_loc:
        m.scatter(lon, lat, latlon=True, marker='*', c='black')
    cb = m.colorbar(matplotlib.cm.ScalarMappable(norm=norm_cb, cmap='seismic'), pad='15%')
    cb.set_label(r'$\mathit{J} \  (mA/m)$')
    ax_cb = cb.ax
    text = ax_cb.yaxis.label
    font_cb = matplotlib.font_manager.FontProperties(family='times new roman', style='italic', size=20)
    text.set_font_properties(font_cb)
    plt.title(label='SECS ' + tp, fontsize=20, color="black", pad=20)
    plt.tight_layout()
    plt.savefig(CONFIG['plots_dir'] + 'SECS' + '_' + date_nightshade.strftime('%Y%m%d%H%M%S') + '.jpeg')
    plt.show()
    print('SECS plots completed!')
    return


def make_plots(dtype='EICS', dtime=None, vplot_sized=True, contour_den=100, s_loc=False, quiver_scale=30, pred = False, pred_var = None, plot_title_addon = None, colormaplimit=None, EICS_component='Jy'):  # or SECS
    """
        This wrapper function to plot the vector and contour map for SECS/EICS data

        Parameters
        ----------
            dtime: str
                Set the single time point of interest with the format
                'YYYY-MM-DD/hh:mm:ss'

            dtype: str
                Data type; valid options: 'EICS' or 'SECS'

            vplot_sized: bool
                Set this flag to plot the EICS' vector field for Jx, Jy with the length for the J's magnitude.

            contour_den: int
                Set the density of the contour map.

            s_loc: bool
                Set this flag to superimpose the virtual stations on the contour map.

        Returns
        ----------
            None.
        """
    if dtype == 'EICS':
        _make_EICS_plots(dtime=dtime, vplot_sized=vplot_sized, contour_den=contour_den, s_loc=s_loc, quiver_scale=quiver_scale, pred = pred, pred_var = pred_var, plot_title_addon = plot_title_addon, colormaplimit=colormaplimit, EICS_component=EICS_component)
        # make a vector map and a contour map.
    if dtype == 'SECS':
        _make_SECS_plots(dtime=dtime, contour_den=contour_den, s_loc=s_loc)
        # make a contour map.
    return


import numpy as np

def lingradest(Bx1, Bx2, Bx3, Bx4,
               By1, By2, By3, By4,
               Bz1, Bz2, Bz3, Bz4,
               R1, R2, R3, R4):
    """
    Calculate magnetic field gradients, divergence, curl, and field line
    curvature from 4-point observations

    Input
    ------
        B-field components: np.ndarrays
            Components of the B-field from the four probes

        Coordinates: np.ndarrays
            Position vectors for the four probes

    Method used
    ------------
        Linear Gradient/Curl Estimator technique
        see Chanteur, ISSI, 1998, Ch. 11

    Notes
    ------
        Based on the IDL version (lingradest.pro), which was
        originally designed for Cluster by A. Runov (2003)

    Returns
    --------
        Dict containing:
            Bxbc, Bybc, Bzbc, Bbc
            LGBx, LGBy, LGBz,
            LCxB, LCyB, LCzB, LD,
            curv_x_B, curv_y_B, curv_z_B, RcurvB
    """

    datarrLength = len(Bx1)

    if len(Bx1) != datarrLength or len(Bx2) != datarrLength or len(Bx3) != datarrLength or len(Bx4) != datarrLength or \
        len(By1) != datarrLength or len(By2) != datarrLength or len(By3) != datarrLength or len(By4) != datarrLength or \
        len(Bz1) != datarrLength or len(Bz2) != datarrLength or len(Bz3) != datarrLength or len(Bz4) != datarrLength or \
        R1.shape[0] != datarrLength or R2.shape[0] != datarrLength or R3.shape[0] != datarrLength or R4.shape[0] != datarrLength:
            print('Problem with input sizes; all input data should be interpolated to the same time stamps')
            return

    Rb = np.zeros((datarrLength, 3))
    dR1 = np.zeros((datarrLength, 3))
    dR2 = np.zeros((datarrLength, 3))
    dR3 = np.zeros((datarrLength, 3))
    dR4 = np.zeros((datarrLength, 3))
    k1 = np.zeros((datarrLength, 3))
    k2 = np.zeros((datarrLength, 3))
    k3 = np.zeros((datarrLength, 3))
    k4 = np.zeros((datarrLength, 3))
    mu1 = np.zeros(datarrLength)
    mu2 = np.zeros(datarrLength)
    mu3 = np.zeros(datarrLength)
    mu4 = np.zeros(datarrLength)
    Bxbc = np.zeros(datarrLength)
    Bybc = np.zeros(datarrLength)
    Bzbc = np.zeros(datarrLength)
    Bbc = np.zeros(datarrLength)
    LGBx = np.zeros((datarrLength, 3)) # Linear Gradient B estimator
    LGBy = np.zeros((datarrLength, 3))
    LGBz = np.zeros((datarrLength, 3))
    LCxB = np.zeros(datarrLength) # Linear Curl B estimator
    LCyB = np.zeros(datarrLength)
    LCzB = np.zeros(datarrLength)
    LD = np.zeros(datarrLength)
    curv_x_B = np.zeros(datarrLength)
    curv_y_B = np.zeros(datarrLength)
    curv_z_B = np.zeros(datarrLength)
    curvB = np.zeros(datarrLength)
    RcurvB = np.zeros(datarrLength)
    B_cross_R_x = np.zeros(datarrLength)
    B_cross_R_y = np.zeros(datarrLength)
    B_cross_R_z = np.zeros(datarrLength)
    B_cross_R = np.zeros(datarrLength)
    Ncurv_x = np.zeros(datarrLength)
    Ncurv_y = np.zeros(datarrLength)
    Ncurv_z = np.zeros(datarrLength)

    # distances in 1000 km!
    r12 = (R2-R1)/1000.0
    r13 = (R3-R1)/1000.0
    r14 = (R4-R1)/1000.0
    r21 = (R1-R2)/1000.0
    r23 = (R3-R2)/1000.0
    r24 = (R4-R2)/1000.0
    r31 = (R1-R3)/1000.0
    r32 = (R2-R3)/1000.0
    r34 = (R4-R3)/1000.0
    r41 = (R1-R4)/1000.0
    r42 = (R2-R4)/1000.0
    r43 = (R3-R4)/1000.0

    for i in range(datarrLength):
        # Tetrahedrom mesocentre coordinates
        Rb[i, 0] = 0.25 * (R1[i, 0] + R2[i, 0] + R3[i, 0] + R4[i, 0])
        Rb[i, 1] = 0.25 * (R1[i, 1] + R2[i, 1] + R3[i, 1] + R4[i, 1])
        Rb[i, 2] = 0.25 * (R1[i, 2] + R2[i, 2] + R3[i, 2] + R4[i, 2])

        # Difference in 1000 km!
        dR1[i, 0:3] = (Rb[i, 0:3] - R1[i, 0:3]) / 1000.0
        dR2[i, 0:3] = (Rb[i, 0:3] - R2[i, 0:3]) / 1000.0
        dR3[i, 0:3] = (Rb[i, 0:3] - R3[i, 0:3]) / 1000.0
        dR4[i, 0:3] = (Rb[i, 0:3] - R4[i, 0:3]) / 1000.0

        k1[i, 0:3] = np.cross(r23[i, 0:3], r24[i, 0:3])
        k1[i, 0:3] = k1[i, 0:3] / (r21[i, 0] * k1[i, 0] + r21[i, 1] * k1[i, 1] + r21[i, 2] * k1[i, 2])
        k2[i, 0:3] = np.cross(r34[i, 0:3], r31[i, 0:3])
        k2[i, 0:3] = k2[i, 0:3] / (r32[i, 0] * k2[i, 0] + r32[i, 1] * k2[i, 1] + r32[i, 2] * k2[i, 2])
        k3[i, 0:3] = np.cross(r41[i, 0:3], r42[i, 0:3])
        k3[i, 0:3] = k3[i, 0:3] / (r43[i, 0] * k3[i, 0] + r43[i, 1] * k3[i, 1] + r43[i, 2] * k3[i, 2])
        k4[i, 0:3] = np.cross(r12[i, 0:3], r13[i, 0:3])
        k4[i, 0:3] = k4[i, 0:3] / (r14[i, 0] * k4[i, 0] + r14[i, 1] * k4[i, 1] + r14[i, 2] * k4[i, 2])

        mu1[i] = 1. + (k1[i, 0] * dR1[i, 0] + k1[i, 1] * dR1[i, 1] + k1[i, 2] * dR1[i, 2])
        mu2[i] = 1. + (k2[i, 0] * dR2[i, 0] + k2[i, 1] * dR2[i, 1] + k2[i, 2] * dR2[i, 2])
        mu3[i] = 1. + (k3[i, 0] * dR3[i, 0] + k3[i, 1] * dR3[i, 1] + k3[i, 2] * dR3[i, 2])
        mu4[i] = 1. + (k4[i, 0] * dR4[i, 0] + k4[i, 1] * dR4[i, 1] + k4[i, 2] * dR4[i, 2])

        # Magnetic field in the barycentre
        Bxbc[i] = mu1[i] * Bx1[i] + mu2[i] * Bx2[i] + mu3[i] * Bx3[i] + mu4[i] * Bx4[i]
        Bybc[i] = mu1[i] * By1[i] + mu2[i] * By2[i] + mu3[i] * By3[i] + mu4[i] * By4[i]
        Bzbc[i] = mu1[i] * Bz1[i] + mu2[i] * Bz2[i] + mu3[i] * Bz3[i] + mu4[i] * Bz4[i]
        Bbc[i] = np.sqrt(Bxbc[i]**2 + Bybc[i]**2 + Bzbc[i]**2)

        LGBx[i, 0:3] = Bx1[i] * k1[i, 0:3] + Bx2[i] * k2[i, 0:3] + Bx3[i] * k3[i, 0:3] + Bx4[i] * k4[i, 0:3]
        LGBy[i, 0:3] = By1[i] * k1[i, 0:3] + By2[i] * k2[i, 0:3] + By3[i] * k3[i, 0:3] + By4[i] * k4[i, 0:3]
        LGBz[i, 0:3] = Bz1[i] * k1[i, 0:3] + Bz2[i] * k2[i, 0:3] + Bz3[i] * k3[i, 0:3] + Bz4[i] * k4[i, 0:3]

        # Divergence B
        LD[i] = Bx1[i] * k1[i, 0] + By1[i] * k1[i, 1] + Bz1[i] * k1[i, 2] + \
                Bx2[i] * k2[i, 0] + By2[i] * k2[i, 1] + Bz2[i] * k2[i, 2] + \
                Bx3[i] * k3[i, 0] + By3[i] * k3[i, 1] + Bz3[i] * k3[i, 2] + \
                Bx4[i] * k4[i, 0] + By4[i] * k4[i, 1] + Bz4[i] * k4[i, 2]

        LCxB[i] = (k1[i, 1] * Bz1[i] - k1[i, 2] * By1[i]) + (k2[i, 1] * Bz2[i] - k2[i, 2] * By2[i]) + \
                  (k3[i, 1] * Bz3[i] - k3[i, 2] * By3[i]) + (k4[i, 1] * Bz4[i] - k4[i, 2] * By4[i])
        LCyB[i] = (k1[i, 2] * Bx1[i] - k1[i, 0] * Bz1[i]) + (k2[i, 2] * Bx2[i] - k2[i, 0] * Bz2[i]) + \
                  (k3[i, 2] * Bx3[i] - k3[i, 0] * Bz3[i]) + (k4[i, 2] * Bx4[i] - k4[i, 0] * Bz4[i])
        LCzB[i] = (k1[i, 0] * By1[i] - k1[i, 1] * Bx1[i]) + (k2[i, 0] * By2[i] - k2[i, 1] * Bx2[i]) + \
                  (k3[i, 0] * By3[i] - k3[i, 1] * Bx3[i]) + (k4[i, 0] * By4[i] - k4[i, 1] * Bx4[i])

        curv_x_B[i] = (Bxbc[i] * LGBx[i, 0] + Bybc[i] * LGBx[i, 1] + Bzbc[i] * LGBx[i, 2]) / (Bbc[i] * Bbc[i])
        curv_y_B[i] = (Bxbc[i] * LGBy[i, 0] + Bybc[i] * LGBy[i, 1] + Bzbc[i] * LGBy[i, 2]) / (Bbc[i] * Bbc[i])
        curv_z_B[i] = (Bxbc[i] * LGBz[i, 0] + Bybc[i] * LGBz[i, 1] + Bzbc[i] * LGBz[i, 2]) / (Bbc[i] * Bbc[i])

        curvB[i] = np.sqrt(curv_x_B[i] * curv_x_B[i] + curv_y_B[i] * curv_y_B[i] + curv_z_B[i] * curv_z_B[i])

        RcurvB[i] = curvB[i]**(-1)

    print('Calculations completed')

    return {'Bxbc': Bxbc, 'Bybc': Bybc, 'Bzbc': Bzbc, 'Bbc': Bbc,
            'LGBx': LGBx, 'LGBy': LGBy, 'LGBz': LGBz,
            'LCxB': LCxB, 'LCyB': LCyB, 'LCzB': LCzB, 'LD': LD,
            'curv_x_B': curv_x_B, 'curv_y_B': curv_y_B, 'curv_z_B': curv_z_B, 'RcurvB': RcurvB}



import logging
import numpy as np
from pytplot import get_data

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)

def mms_get_hpca_dist(tname, index=None, probe=None, data_rate=None, species=None, times=False):
    """
    Returns 3D particle data structures containing MMS HPCA
    data for use with SPEDAS particle routines. 
    
    Input
    ----------
        tname: str
            tplot variable name containing the HPCA distribution data

    Parameters
    ----------
        probe: str
            Spacecraft probe #

        data_rate: str
            Instrument data rates for HPCA include 'brst' and 'srvy'

        species: str
            Species of the data specified by the 'tname' input

        index: int
            Index of time sample to return

        times: bool
            Flag to return the distribution times instead of the particle data structure(s)

    Returns:
        3D particle data structure(s) containing MMS HPCA distribution functions
    """

    data_in = get_data(tname)

    # Match particle data to azimuth data
    # ------------------------

    # get azimuths and full dist sample times from ancillary variable
    azimuth = get_data('mms' + probe + '_hpca_azimuth_angles_per_ev_degrees')

    if azimuth is None:
        logging.error('No azimuth data found for the current time range')
        return

    # check if the time series is monotonic to avoid doing incorrect 
    # calculations when there's a problem with the CDF files
    time_data = azimuth.times

    wherenonmono = np.where(time_data[1:] <= time_data[:-1])
    if len(wherenonmono[0]) != 0:
        logging.error('Error, non-monotonic data found in the HPCA Epoch_Angles time series data')
        return

    # find azimuth times with complete 1/2 spins of particle data
    # this is used to determine the number of 3D distributions that will be created
    # and where their corresponding data is located in the particle data structure
    n_times = len(azimuth.y[0, 0, :])
    data_idx = np.searchsorted(data_in.times, time_data)-1
    full = np.argwhere((data_idx[1:]-data_idx[:-1]) == n_times)
    if len(full) == 0:
        logging.error('Azimuth data does not cover current data\'s time range')
        return


    # filter times when azimuth data is all zero
    #   -just check the first energy & elevation
    #   -assume azimuth values are positive
    n_valid_az = np.count_nonzero(azimuth.y[full, 0, 0, :])
    if n_valid_az != azimuth.y[full, 0, 0, :].size:
        logging.error('Error, zeroes found in the azimuth array')
        return
        # should check to see if all azimuths are 0 (and if not, use the valid ones)

    if index is not None:
        if index >= full.size:
            logging.error('Error, index greater than data size')
            return
        full = full[index]

    full = full.squeeze()

    data_idx = data_idx[full].flatten()

    if times:
        return time_data[full]

    # Initialize energies, angles, and support data
    # ------------------------
    
    # final dimensions for a single distribution (energy-azimuth-elevation)
    azimuth_dim = azimuth.y.shape
    dim = (azimuth_dim[3], azimuth_dim[2], azimuth_dim[1])

    if species == 'hplus':
        mass = 1.04535e-2
        charge = 1.0
    elif species == 'heplus':
        mass = 4.18138e-2
        charge = 1.0
    elif species == 'heplusplus':
        mass = 4.18138e-2
        charge = 2.0
    elif species == 'oplus':
        mass = 0.167255
        charge = 1.0
    elif species == 'oplusplus':
        mass = 0.167255
        charge = 2.0
    else:
        logging.error('Cannot determine species')
        return

    out = {'project_name': 'MMS',
           'spacecraft': probe, 
           'species': species,
           'data_name': 'HPCA ' + species,
           'charge': charge,
           'units_name': 'df_cm',
           'mass': mass}

    out_bins = np.zeros(dim) + 1
    out_denergy = np.zeros(dim)

    # energy bins are constant
    energy_reform = np.reshape(data_in.v2, [len(data_in.v2), 1, 1])
    energy_rebin1 = np.repeat(energy_reform, dim[1], axis=2) # repeated across theta
    out_energy = np.repeat(energy_rebin1, dim[2], axis=1)
    energy_len = len(data_in.v2)

    # elevations are constant across time
    # convert colat -> lat
    theta_reform = 90. - np.reshape(data_in.v1, [1, 1, len(data_in.v1)])

    # in the IDL code, we use reform to repeat the vector above
    # here, we'll do the same thing with np.repeat
    theta_rebin1 = np.repeat(theta_reform, azimuth_dim[1], axis=1) # repeated across phi
    out_theta = np.repeat(theta_rebin1, data_in[1].shape[2], axis=0) # repeated across energy
    theta_len = len(data_in.v1)
    out_dtheta = np.zeros([energy_len, theta_len, azimuth_dim[2]]) + 22.5

    # get start/end times
    # - this assumes that the times from the particle (and angle) data
    #   are at the center of the corresponding energy sweep
    # - also assumes that there are no gaps in the data
    dt = time_data[1:]-time_data[0:-1]
    dt_sweep = data_in.times[1:]-data_in.times[0:-1]
    start_times = time_data[full]-dt_sweep[data_idx]
    end_times = start_times + dt[full]

    # get azimuth 
    #  -shift from time-azimuth-elevation-energy to time-energy-azimuth-elevation
    out_phi = azimuth.y[full, :, :, :]

    if len(out_phi.shape) == 4:
        out_phi = out_phi.transpose([0, 3, 1, 2])
    elif len(out_phi.shape) == 3:
        out_phi = out_phi.transpose([2, 0, 1])
    
    phi_len = azimuth_dim[1]

    # get dphi
    #  -use median distance between subsequent phi measurements within each distribution
    #   (median is used to discard large differences across 0=360)
    #  -preserve dimensionality in case differences arise across energy or elevation
    
    if len(out_phi.shape) == 4:
        out_dphi = np.median(azimuth.y[full, 1:, :, :] - azimuth.y[full, :-1, :, :], axis=1)
        out_dphi = out_dphi.transpose([0, 2, 1])
    elif len(out_phi.shape) == 3:
        out_dphi = np.median(azimuth.y[full, 1:, :, :] - azimuth.y[full, :-1, :, :], axis=0)
        out_dphi = out_dphi.transpose([1, 0])

    if len(out_phi.shape) == 4:
        dphi_reform = np.reshape(out_dphi, [full.size, energy_len, theta_len, 1])
        out_dphi = np.repeat(dphi_reform, phi_len, axis=3)
    elif len(out_phi.shape) == 3:
        dphi_reform = np.reshape(out_dphi, [energy_len, theta_len, 1])
        out_dphi = np.repeat(dphi_reform, phi_len, axis=2)

    out_data = np.zeros((full.size, dim[0], dim[1], dim[2]))

    # copy particle data
    for i in range(full.size):
        # need to extract the data from the center of the half-spin
        if data_idx[i]-n_times/2.0 < 0:
            start_idx = 0
        else:
            start_idx = int(data_idx[i]-n_times/2.)

        if data_idx[i]+n_times/2.-1 >= len(data_in.times):
            end_idx = len(data_in.times)
        else:
            end_idx = int(data_idx[i]+n_times/2.)

        out_data[i, :, :, :] = data_in.y[start_idx:end_idx, :, :].transpose([2, 0, 1])

    out_list = []

    for time_idx, time in enumerate(start_times):
        out_table = {**out}
        out_table['data'] = out_data[time_idx, :, :, :]
        out_table['bins'] = out_bins
        out_table['theta'] = out_theta
        if len(out_phi.shape) == 4:
            out_table['phi'] = out_phi[time_idx, :, :, :]
        else:
            out_table['phi'] = out_phi
        out_table['energy'] = out_energy
        out_table['dtheta'] = out_dtheta
        if len(out_phi.shape) == 4:
            out_table['dphi'] = out_dphi[time_idx, :, :, :]
        else:
            out_table['dphi'] = out_dphi
        out_table['denergy'] = out_denergy
        out_table['n_energy'] = energy_len
        out_table['n_theta'] = len(data_in[2])
        out_table['n_phi'] = phi_len
        out_table['start_time'] = time
        out_table['end_time'] = end_times[time_idx]
        out_list.append(out_table)

    return out_list









import os
import unittest
from pyspedas.utilities.data_exists import data_exists

import pyspedas
from pytplot import del_data

class LoadTestCases(unittest.TestCase):
    def tearDown(self):
        del_data('*')

    def test_downloadonly(self):
        files = pyspedas.rbsp.efw(trange=['2015-11-3', '2015-11-4'], level='l3', downloadonly=True)
        self.assertTrue(os.path.exists(files[0]))

    # temporarily disabled, 3Dec2021
    # def test_notplot(self):
    #     data = pyspedas.rbsp.efw(trange=['2015-11-6', '2015-11-7'], level='l3', notplot=True)
    #     self.assertTrue('density' in data.keys())
    #     self.assertTrue('Vavg' in data.keys())
    #     self.assertTrue('vel_gse' in data.keys())
    #     self.assertTrue('efield_inertial_frame_mgse' in data.keys())
    #     self.assertTrue('x' in data['density'].keys())
    #     self.assertTrue('y' in data['density'].keys())

    def test_load_emfisis_data(self):
        emfisis_vars = pyspedas.rbsp.emfisis(trange=['2018-11-5', '2018-11-6'], datatype='magnetometer', level='l3', time_clip=True)
        wfr_vars = pyspedas.rbsp.emfisis(trange=['2018-11-5', '2018-11-6'], level='l2', datatype='wfr')
        hfr_vars = pyspedas.rbsp.emfisis(trange=['2018-11-5', '2018-11-6'], level='l2', datatype='hfr')
        self.assertTrue(data_exists('Mag'))
        # WFR waveform data
        self.assertTrue(data_exists('BuSamples'))
        self.assertTrue(data_exists('BvSamples'))
        self.assertTrue(data_exists('BwSamples'))
        self.assertTrue(data_exists('EuSamples'))
        self.assertTrue(data_exists('EvSamples'))
        self.assertTrue(data_exists('EwSamples'))
        # HFR waveform data
        self.assertTrue(data_exists('HFRsamples'))

    def test_load_efw_data(self):
        efw_vars = pyspedas.rbsp.efw(trange=['2015-11-5', '2015-11-6'], level='l3')
        self.assertTrue(data_exists('density'))

    # def test_load_rbspice_data(self):
    #     rbspice_vars = pyspedas.rbsp.rbspice(trange=['2018-11-5', '2018-11-6'], datatype='tofxeh', level='l3')
    #     self.assertTrue(data_exists('Alpha'))

    def test_load_mageis_data(self):
        mageis_vars = pyspedas.rbsp.mageis(trange=['2018-11-5', '2018-11-6'], level='l3', rel='rel04')
        self.assertTrue(data_exists('I'))

    def test_load_hope_data(self):
        hope_vars = pyspedas.rbsp.hope(trange=['2018-11-5', '2018-11-6'], datatype='moments', level='l3', rel='rel04')
        self.assertTrue(data_exists('Ion_density'))

    def test_load_rep_data(self):
        rept_vars = pyspedas.rbsp.rept(trange=['2018-11-4', '2018-11-5'], level='l2', rel='rel03')
        self.assertTrue(data_exists('FESA'))
        self.assertTrue(data_exists('FEDU'))
        self.assertTrue(data_exists('FPSA'))
        self.assertTrue(data_exists('FPDU'))

    # disabled, 16Sep2021 due to a problem with the data
    # def test_load_rps1min_data(self):
    #   rps_vars = pyspedas.rbsp.rps()
    #   self.assertTrue(data_exists('DOSE2_RATE'))

if __name__ == '__main__':
    unittest.main()


import numpy as np

from .slice2d_intrange import slice2d_intrange
from .slice2d_get_data import slice2d_get_data
from .slice2d_geo import slice2d_geo
from .slice2d_get_support import slice2d_get_support
from .slice2d_orientslice import slice2d_orientslice
from .slice2d_rotate import slice2d_rotate
from .slice2d_custom_rotation import slice2d_custom_rotation
from .slice2d_nearest import slice2d_nearest
from .slice2d_rlog import slice2d_rlog
from .slice2d_s2c import slice2d_s2c
from .slice2d_2di import slice2d_2di
from .slice2d_smooth import slice2d_smooth
from .slice2d_subtract import slice2d_subtract

from pyspedas import time_double, time_string


def slice2d(dists,
            time=None,
            samples=None,
            window=None,
            center_time=False,
            interpolation='geometric',
            trange=None,
            resolution=None,
            rotation='xy',
            custom_rotation=None,
            subtract_bulk=False,
            energy=False,
            log=False,
            erange=None,
            smooth=None,
            thetarange=None,
            zdirrange=None,
            mag_data=None,
            vel_data=None,
            sun_data=None,
            slice_x=None,
            slice_z=None):
    """
    Returns an interpolated 2D slice of 3D particle data for plotting

    Interpolation methods:

    2D Interpolation:
        Data points within the specified theta or z-axis range are projected onto
        the slice plane and linearly interpolated onto a regular 2D grid.

    Geometric:
        Each point on the plot is given the value of the bin it intersects.
        This allows bin boundaries to be drawn at high resolutions.

    Input
    ---------------------
        dists: list of dicts
            List of 3D particle data structures

    Basic Keywords
    ---------------------
        trange: list of str or list of float
            Two-element time range over which data will be averaged (optional)

        time: str
            Time at which the slice will be computed (optional)

        samples: int
            Numer of samples nearest to TIME to average (default 1)

        window: int or float
            Length in seconds from TIME over which data will be averaged.

        center_time: bool
            Flag denoting that TIME should be midpoint for window instead of beginning.

        interpolation: str
            Interpolation method to use; options: 'geometric' for geometric interpolation
            and '2d' for 2D interpolation (described above)

    Orientation Keywords
    ---------------------
        custom_rotation: str or np.ndarray
            Applies a custom rotation matrix to the data.  Input may be a
            3x3 rotation matrix or a tplot variable containing matrices.
            If the time window covers multiple matrices they will be averaged.

        rotation: str
            Aligns the data relative to the magnetic field and/or bulk velocity.
            This is applied after the CUSTOM_ROTATION. (BV and BE are invariant
            between coordinate systems)

            Use MAG_DATA keyword to specify magnetic field vector.
            Use VEL_DATA keyword to specify bulk velocity (optional).

            'BV':  The x-axis is parallel to B field; the bulk velocity defines the x-y plane
            'BE':  The x-axis is parallel to B field; the B x V(bulk) vector defines the x-y plane
            'xy':  (default) The x-axis is along the data's x-axis and y is along the data's y axis
            'xz':  The x-axis is along the data's x-axis and y is along the data's z axis
            'yz':  The x-axis is along the data's y-axis and y is along the data's z axis
            'xvel':  The x-axis is along the data's x-axis; the x-y plane is defined by the bulk velocity
            'perp':  The x-axis is the bulk velocity projected onto the plane normal to the B field; y is B x V(bulk)
            'perp_xy':  The data's x & y axes are projected onto the plane normal to the B field
            'perp_xz':  The data's x & z axes are projected onto the plane normal to the B field
            'perp_yz':  The data's y & z axes are projected onto the plane normal to the B field

        mag_data: str
            Name of tplot variable containing magnetic field data or 3-vector.
            This will be used for slice plane alignment and must be in the
            same coordinates as the particle data.

        vel_data: str
            Name of tplot variable containing the bulk velocity data or 3-vector.
            This will be used for slice plane alignment and must be in the
            same coordinates as the particle data.
            If not set the bulk velocity will be automatically calculated
            from the distribution (when needed).

    Other Keywords
    ---------------------
        resolution: int
            Integer specifying the resolution along each dimension of the
            slice (defaults:  2D interpolation: 150, geometric: 500)

        smooth: int
            An odd integer >=3 specifying the width of a smoothing window in #
            of points.  Smoothing is applied to the final plot using a gaussian
            convolution. Even entries will be incremented, 0 and 1 are ignored.

        energy: bool
            Flag to plot data against energy (in eV) instead of velocity.

        log: bool
            Flag to apply logarithmic scaling to the radial measure (i.e. energy/velocity).
            (on by default if energy=True)

        erange: list of float
            Two element list specifying the energy range to be used in eV

        thetarange: list of float
            (2D interpolation only): angle range, in degrees (-90, +90), used to calculate the slice
            default = [-20, 20]; will override zdirrange

        zdirrange: list of float
            (2D interpolation only): Z-Axis range, in km/s, used to calculate slice.
            Ignored if called with THETARANGE.

        subtract_bulk: bool
            Flag to subtract the bulk velocity vector

    Returns
    ---------------------
        Dictionary containing 2D slice of 3D particle data
    """

    if trange is None:
        if time is None:
            print('Please specify a time or time range over which to compute the slice.')
            return
        if window is None and samples is None:
            # use single closest distribution by default
            samples = 1

    valid_rotations = ['bv', 'be', 'xy', 'xz', 'yz', 'xvel', 'perp', 'perp2',
                       'perp_xy', 'perp_xz', 'perp_yz', 'b_exb', 'perp1-perp2']

    if rotation not in valid_rotations:
        print('Invalid rotation requested; valid options: ' + ', '.join(valid_rotations))
        return

    if interpolation == '2d':
        if smooth is None:
            smooth = 7
        if resolution is None:
            resolution = 150
        if thetarange is None and zdirrange is None:
            thetarange = [-20, 20]
    elif interpolation == 'geometric':
        if resolution is None:
            resolution = 500

    if time is not None:
        time = time_double(time)

    # always use log scale when energy is requested instead of velocity
    if energy and not log:
        log = True

    # Get the slice's time range
    # ------------------------------------------------------------------------
    # get the time range if a time & window were specified instead
    if trange is None and window is not None:
        if center_time:
            trange = [time - window/2.0, time + window/2.0]
        else:
            trange = [time, time + window]

    # if no trange or window was specified, then get a time range
    # from the N closest samples to the specified time
    # (defaults to 1 if samples is not defined)
    if trange is None:
        trange = slice2d_nearest(dists, time, samples)

    tr = time_double(trange)

    msg_prefix = 'Processing slice at ' + time_string(tr[0], fmt='%Y-%m-%d %H:%M:%S.%f') + '... '

    # check that there is data in range before proceeding
    times_ind = slice2d_intrange(dists, tr)

    # Extract particle data
    #   - apply energy limits
    #   - average data over the time window
    #   - output r, phi, theta and dr, dphi and dtheta arrays
    # ------------------------------------------------------------------------
    data = slice2d_get_data(dists, trange=tr, erange=erange)

    # get original data and radial ranges for plotting
    #   - ignore outliers that may be the result of sanitizations performed outside this routine
    #   - this also effectively removes low-significance outliers at large velocities, which
    #     can lead to overly large velocity scales for some instruments
    idx = np.argwhere(data['data'] > 0.0)
    if len(idx) > 0:
        dmean = np.mean(np.log10(data['data'][idx]))
        dvar = np.var(np.log10(data['data'][idx]))
        dmin = 10**(dmean - 2*np.sqrt(dvar))
        drange = [dmin, np.nanmax(data['data'][idx])]
    else:
        drange = [0.0, 0.0]

    rrange = [np.nanmin(data['rad'] - 0.5*data['dr']), np.nanmax(data['rad'] + 0.5*data['dr'])]

    # apply radial log scaling
    if log:
        scaled_r = slice2d_rlog(data['rad'], data['dr'])
        # replace original vars
        data['rad'] = scaled_r['r']
        data['dr'] = scaled_r['dr']

    # convert spherical data to cartesian coordinates for interpolation
    xyz = None
    if interpolation != 'geometric':
        xyz = slice2d_s2c(data['rad'], data['theta'], data['phi'])

    # Get/apply coordinate transformations
    #  - for 2D/3D interpolation, the data is transformed here
    #  - for geometric interpolation, the data is transformed in slice2d_geo
    #  - support data is always rotated at each step
    # ------------------------------------------------------------------------
    # get support data for aligning slice
    bfield = slice2d_get_support(mag_data, trange)
    vbulk = slice2d_get_support(vel_data, trange)
    sunvec = slice2d_get_support(sun_data, trange)
    slice_x_vec = slice2d_get_support(slice_x, trange)
    slice_z_vec = slice2d_get_support(slice_z, trange)

    # custom rotation
    custom_rot = slice2d_custom_rotation(custom_rotation=custom_rotation,
                                         trange=trange,
                                         vectors=xyz,
                                         bfield=bfield,
                                         vbulk=vbulk,
                                         sunvec=sunvec)

    # for 2D/3D interpolation
    orientation = slice2d_orientslice(vectors=custom_rot['vectors'],
                                      vbulk=custom_rot['vbulk'],
                                      bfield=custom_rot['bfield'],
                                      sunvec=custom_rot['sunvec'],
                                      slice_x=slice_x_vec,
                                      slice_z=slice_z_vec)

    # built-in rotation option
    rot_matrix = slice2d_rotate(rotation=rotation,
                                vectors=orientation['vectors'],
                                bfield=orientation['bfield'],
                                vbulk=orientation['vbulk'],
                                sunvec=orientation['sunvec'])

    # problem rotating data
    if rot_matrix is None:
        return

    geo_shift = [0, 0, 0]
    if subtract_bulk:
        vectors = slice2d_subtract(rot_matrix['vectors'], vbulk)

        if vectors is not None:
            rot_matrix['vectors'] = vectors

        # the shift is applied later for geometric interpolation
        geo_shift = vbulk

    # sort transformed vector grid
    if xyz is not None:
        sorted_idxs = np.argsort(rot_matrix['vectors'][:, 0], kind='stable')
        rot_matrix['vectors'] = rot_matrix['vectors'][sorted_idxs, :]
        data['data'] = data['data'][sorted_idxs]

    # Create the slice
    # ------------------------------------------------------------------------
    if interpolation == 'geometric':
        the_slice = slice2d_geo(data['data'], resolution, data['rad'], data['phi'], data['theta'], data['dr'], data['dp'],
                          data['dt'], orient_matrix=orientation['matrix'], rotation_matrix=rot_matrix['matrix'],
                          custom_matrix=custom_rot['matrix'], msg_prefix=msg_prefix, shift=geo_shift)
    elif interpolation == '2d':
        the_slice = slice2d_2di(data['data'], rot_matrix['vectors'], resolution, thetarange=thetarange, zdirrange=zdirrange)

    if smooth is not None:
        the_slice = slice2d_smooth(the_slice, smooth)

    # Get metadata and return slice
    # ------------------------------------------------------------------------
    if energy:
        xyunits = 'eV'
    else:
        xyunits = 'km/s'

    out = {'project_name': dists[0]['project_name'],
           'spacecraft': dists[0]['spacecraft'],
           'data_name': dists[0]['data_name'],
           'units_name': dists[0]['units_name'],
           'species': dists[0]['species'],
           'xyunits': xyunits,
           'rotation': rotation,
           'energy': energy,
           'trange': trange,
           'zrange': drange,
           'rrange': rrange,
           'rlog': log,
           'n_samples': len(times_ind),
           **the_slice}

    print('Finished slice at ' + time_string(tr[0], fmt='%Y-%m-%d %H:%M:%S.%f'))
    return out




import requests

from pyspedas.themis.load import load


def gmag(trange=['2007-03-23', '2007-03-24'],
         sites=None,
         group=None,
         level='l2',
         suffix='',
         get_support_data=False,
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False):
    """
    This function loads ground magnetometer data

    Parameters:
        trange: list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        level: str
            Data type; Valid options: 'l1', 'l2'

        suffix: str
            The tplot variable names will be given this suffix.
            By default, no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load
            (if not specified, all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified
            in the trange keyword

        sites: str/list of str
            GMAG station names to load (e.g. 'bmls').

        group: str
            GMAG group of stations (eg. 'epo').
            If specified, stations is ignored.

    Returns:
        List of tplot variables created.

    """

    if sites is None:
        thm_sites = 'atha chbg ekat fsim fsmi fykn gbay glyn gill inuv kapu '\
                     'kian kuuj mcgr nrsq pgeo pina rank snap snkq tpas whit '\
                     'yknf'.split(' ')
        tgo_sites = ['nal', 'lyr', 'hop', 'bjn', 'nor', 'sor', 'tro', 'and',
                     'don', 'rvk', 'sol', 'kar', 'jan', 'jck', 'dob']
        dtu_sites = ['atu', 'dmh', 'svs', 'tdc', 'bfe', 'roe', 'thl', 'kuv',
                     'upn', 'umq', 'gdh', 'stf', 'skt', 'ghb', 'fhb', 'naq',
                     'amk', 'sco', 'tab', 'sum', 'hov']
        ua_sites = ['arct', 'bett', 'cigo', 'eagl', 'fykn', 'gako', 'hlms',
                    'homr', 'kako', 'pokr', 'trap']
        maccs_sites = ['cdrt', 'chbr', 'crvr', 'gjoa', 'iglo', 'nain',
                       'pang', 'rbay']
        usgs_sites = ['bou', 'brw', 'bsl', 'cmo', 'ded', 'frd', 'frn',
                      'gua', 'hon', 'new', 'shu', 'sit', 'sjg', 'tuc']
        atha_sites = ['roth', 'leth', 'redr', 'larg', 'vldr', 'salu', 'akul',
                      'puvr', 'inuk', 'kjpk', 'radi', 'stfl', 'sept', 'schf']
        epo_sites = ['bmls', 'ccnv', 'drby', 'fyts', 'hots', 'loys',
                     'pgeo', 'pine', 'ptrs', 'rmus', 'swno', 'ukia']
        falcon_sites = ['hris', 'kodk', 'lrel', 'pblo', 'stfd', 'wlps']
        mcmac_sites = ['amer', 'benn', 'glyn', 'lyfd', 'pcel', 'rich',
                       'satx', 'wrth']
        nrcan_sites = ['blc', 'cbb', 'iqa', 'mea', 'ott', 'stj', 'vic']
        step_sites = ['fsj', 'ftn', 'hrp', 'lcl', 'lrg', 'pks', 'whs']
        fmi_sites = ['han', 'iva', 'kev', 'kil', 'mas', 'mek', 'muo', 'nur',
                     'ouj', 'pel', 'ran', 'tar']
        aair_sites = ['amd', 'bbg', 'brn', 'dik', 'loz', 'pbk', 'tik', 'viz']
        carisma_sites = ['anna', 'back', 'cont', 'daws', 'eski', 'fchp',
                         'fchu', 'gull', 'isll', 'lgrr', 'mcmu', 'mstk',
                         'norm', 'osak', 'oxfo', 'pols', 'rabb', 'sach',
                         'talo', 'thrf', 'vulc', 'weyb', 'wgry']
        sites = (thm_sites + tgo_sites + dtu_sites + ua_sites + maccs_sites
                 + usgs_sites + atha_sites + epo_sites + falcon_sites
                 + mcmac_sites + nrcan_sites + step_sites + fmi_sites
                 + aair_sites + carisma_sites)

    if group is not None:
        sites = eval(group+'_sites')

    if not isinstance(sites, list):
        sites = [sites]

    # check for sites in Greenland
    greenland = []
    for site in sites:
        if check_greenland(site):
            greenland.append(True)
        else:
            greenland.append(False)

    return load(instrument='gmag', trange=trange, level=level,
                suffix=suffix, get_support_data=get_support_data,
                varformat=varformat, varnames=varnames,
                downloadonly=downloadonly, notplot=notplot, stations=sites,
                greenland=greenland, time_clip=time_clip, no_update=no_update)


gmag_dict = {}


def query_gmags():
    """ returns a dictionary of gmag stations and all their metadata
    """

    url = 'http://themis.ssl.berkeley.edu/gmag/gmag_json.php'
    global gmag_dict

    params = dict(
        station='',
        group=''
    )
    resp = requests.get(url=url, params=params)
    data = resp.json()
    gmag_dict = data

    return data


def get_group(station_name):
    """ returns a list with the groups that station belongs to
    """

    global gmag_dict
    group_list = []
    if gmag_dict == {}:
        gmag_dict = query_gmags()

    for station in gmag_dict:
        if station['ccode'].lower() == station_name:
            for key, value in station.items():
                if value == 'Y':
                    if station['ccode'].lower() not in group_list:
                        group_list.append(key)

    return group_list


def gmag_list(group='all'):
    """ returns a list of stations
        prints a list of "stations: start date - end date"
    """

    global gmag_dict
    station_list = []
    if gmag_dict == {}:
        gmag_dict = query_gmags()

    for station in gmag_dict:
        station_name = station['ccode'].lower()
        station_group = get_group(station_name)
        if group in ['all', '*', ''] or group in station_group:
            station_list.append(station_name)
            print(station_name + ": from " + station['day_first'] + " to "
                  + station['day_last'])

    return station_list


def gmag_groups():
    """ returns a dictionary of station groups with a list of stations
        prints a list of "group:'stations'"
    """

    global gmag_dict
    group_dict = {}
    if gmag_dict == {}:
        gmag_dict = query_gmags()

    for station in gmag_dict:
        for key, value in station.items():
            if value == 'Y':
                if key in group_dict:
                    if station['ccode'].lower() not in group_dict[key]:
                        group_dict[key].append(station['ccode'].lower())
                else:
                    group_dict[key] = []
                    group_dict[key].append(station['ccode'].lower())

    # print them
    print()
    for g, s in group_dict.items():
        print(g + ":" + ",'".join(s) + "'")

    return group_dict


def check_gmag(station_name):
    """ returns 1 if station_name is in the gmag list, 0 otherwise"
    """

    global gmag_dict
    if gmag_dict == {}:
        gmag_dict = query_gmags()

    for station in gmag_dict:
        if station['ccode'].lower() == station_name.lower():
            return 1

    return 0


def check_greenland(station_name):
    """ returns 1 if station_name is in the greenland gmag list, 0 otherwise"
    """

    global gmag_dict
    if gmag_dict == {}:
        gmag_dict = query_gmags()

    for station in gmag_dict:
        if station['ccode'].lower() == station_name.lower():
            if ((station['country'] is not None
                 and station['country'].lower() == 'greenland')
                or (station['greenland'] is not None
                    and station['greenland'].lower() == 'y')):
                return 1

    return 0



import unittest
from pyspedas.mms.particles.mms_part_slice2d import mms_part_slice2d


class SliceTests(unittest.TestCase):
    def test_fpi_brst_rotations(self):
        time = '2015-10-16/13:06:30'
        rotations = ['xy', 'xz', 'bv', 'be', 'xvel', 'perp', 'perp_xy', 'perp_xz', 'perp_yz', 'b_exb', 'perp1-perp2']
        species = ['i', 'e']
        for spc in species:
            for rotation in rotations:
                mms_part_slice2d(time=time, probe='1', species=spc, data_rate='brst', rotation=rotation, save_png='test_fpi_brst_' + spc + '_' + rotation, display=False)


if __name__ == '__main__':
    unittest.main()
